package sorting;

import java.util.*;

/**
 * The standard generic list of data
 * @author Viera K. Proulx
 *
 * @param <T> the type of data in this list
 */
interface ILo<T>{
  /** 
   * produce a sorted list from this list
   * sorted in the order given by the <code>Comparator</code>
   * @param comp the given <code>Comparator</code>
   * @return a sorted list
   */
  public ILo<T> quicksort(Comparator<T> comp);
  
  /**
   * produce a list of all items 
   * lower than or equal to the given pivot
   * @param pivot the given pivot
   * @param comp the given <code>Comparator</code>
   * @return a list of item lower than or equal to the pivot
   */
  public ILo<T> lower(T pivot, Comparator<T> comp);

  /**
   * produce a list of all items 
   * greater than the given pivot
   * @param pivot the given pivot
   * @param comp the given <code>Comparator</code>
   * @return a list of item greater than the pivot
   */
  public ILo<T> upper(T pivot, Comparator<T> comp);
  
  /**
   * append the given list to this list
   * @param that the given list
   * @return the two lists appended
   */
  public ILo<T> append(ILo<T> that);

  /**
   * produce a sorted list from this list
   * sorted in the order given by the Comparator
   * using the insertion sort algorithm
   * @param comp the given <code>Comparator</code>
   * @return the sorted list
   */
  public ILo<T> insertionsort(Comparator<T> comp);

  /**
   * insert the given item into this list
   * that is sorted in the order given by the Comparator
   * @param comp the given <code>Comparator</code>
   * @return the sorted list with the given item inserted
   */
  public ILo<T> insert(T item, Comparator<T> comp);
  
  /**
   * Print this list
   */
  public void printlist();
  
  /**
   * Is this list sorted in the order given by the <code>Comparator</code>
   * @param comp the given <code>Comparator</code>
   * @return true if the list is sorted
   */
  public boolean isSorted(Comparator<T> comp);
  
  /**
   * Is this list sorted after the given item 
   * in the order given by the <code>Comparator</code>
   * @param item the given item that should precede the list
   * @param comp the given <code>Comparator</code>
   * @return true if the list is sorted
   */
  public boolean isSortedAfter(T item, Comparator<T> comp);
}

/**
 * The standard generic empty list of data
 * @author Viera K. Proulx
 *
 * @param <T> the type of data in this list
 */
class MtLo<T> implements ILo<T>{
  MtLo(){}
  
  /** 
   * produce a sorted list from this list
   * sorted in the order given by the <code>Comparator</code>
   * @param comp the given <code>Comparator</code>
   * @return a sorted list
   */
  public ILo<T> quicksort(Comparator<T> comp){
    return this;
  }
  
  /**
   * produce a list of all items 
   * lower than or equal to the given pivot
   * @param pivot the given pivot
   * @param comp the given <code>Comparator</code>
   * @return a list of item lower than or equal to the pivot
   */
  public ILo<T> lower(T pivot, Comparator<T> comp){
    return this;
  }

  /**
   * produce a list of all items 
   * greater than the given pivot
   * @param pivot the given pivot
   * @param comp the given <code>Comparator</code>
   * @return a list of item greater than the pivot
   */
  public ILo<T> upper(T pivot, Comparator<T> comp){
    return this;
  }
  
  /**
   * append the given list to this list
   * @param that the given list
   * @return the two lists appended
   */
  public ILo<T> append(ILo<T> that){
    return that;
  }

  /**
   * produce a sorted list from this list
   * sorted in the order given by the Comparator
   * using the insertion sort algorithm
   * @param comp the given <code>Comparator</code>
   * @return the sorted list
   */
  public ILo<T> insertionsort(Comparator<T> comp){
    return this;
  }

  /**
   * insert the given item into this list
   * that is sorted in the order given by the Comparator
   * @param comp the given <code>Comparator</code>
   * @return the sorted list with the given item inserted
   */
  public ILo<T> insert(T item, Comparator<T> comp){
    return new ConsLo<T>(item, this);
  }
  
  /**
   * Print this list
   */
  public void printlist(){
    System.out.println("THE END");
  }
  
  /**
   * Is this list sorted in the order given by the <code>Comparator</code>
   * @param comp the given <code>Comparator</code>
   * @return true if the list is sorted
   */
  public boolean isSorted(Comparator<T> comp){
    return true;
  }
  
  /**
   * Is this list sorted after the given item 
   * in the order given by the <code>Comparator</code>
   * @param item the given item that should precede the list
   * @param comp the given <code>Comparator</code>
   * @return true if the list is sorted
   */
  public boolean isSortedAfter(T item, Comparator<T> comp){
    return true;
  }
}

/**
 * The standard generic cons-list of data
 * @author Viera K. Proulx
 *
 * @param <T> the type of data in this list
 */
class ConsLo<T> implements ILo<T>{
  T first;
  ILo<T> rest;
  
  ConsLo(T first, ILo<T> rest){
    this.first = first;
    this.rest = rest;
  }
  
  /** 
   * produce a sorted list from this list
   * sorted in the order given by the <code>Comparator</code>
   * @param comp the given <code>Comparator</code>
   * @return a sorted list
   */
  public ILo<T> quicksort(Comparator<T> comp){
    return this.rest.lower(this.first, comp).quicksort(comp).
           append(new ConsLo<T>(this.first, 
           this.upper(this.first, comp).quicksort(comp)));
  }
  
  /**
   * produce a list of all items 
   * lower than or equal to the given pivot
   * @param pivot the given pivot
   * @param comp the given <code>Comparator</code>
   * @return a list of item lower than or equal to the pivot
   */
  public ILo<T> lower(T pivot, Comparator<T> comp){
    if (comp.compare(this.first, pivot) <= 0)
      return new ConsLo<T>(this.first, this.rest.lower(pivot, comp));
    else
      return this.rest.lower(pivot, comp);
  }

  /**
   * produce a list of all items 
   * greater than the given pivot
   * @param pivot the given pivot
   * @param comp the given <code>Comparator</code>
   * @return a list of item greater than the pivot
   */
  public ILo<T> upper(T pivot, Comparator<T> comp){
    if (comp.compare(this.first, pivot) > 0)
      return new ConsLo<T>(this.first, this.rest.upper(pivot, comp));
    else
      return this.rest.upper(pivot, comp);
  }
  
  /**
   * append the given list to this list
   * @param that the given list
   * @return the two lists appended
   */
  public ILo<T> append(ILo<T> that){
    return new ConsLo<T>(this.first, this.rest.append(that));
  }

  /**
   * produce a sorted list from this list
   * sorted in the order given by the Comparator
   * using the insertion sort algorithm
   * @param comp the given <code>Comparator</code>
   * @return the sorted list
   */
  public ILo<T> insertionsort(Comparator<T> comp){
    return this.rest.insertionsort(comp).insert(this.first, comp);
  }

  /**
   * insert the given item into this list
   * that is sorted in the order given by the Comparator
   * @param comp the given <code>Comparator</code>
   * @return the sorted list with the given item inserted
   */
  public ILo<T> insert(T item, Comparator<T> comp){
    if (comp.compare(item, this.first) <= 0)
      return new ConsLo<T>(item, this);
    else
      return new ConsLo<T>(this.first, this.rest.insert(item, comp));
  }
    
  /**
   * Print this list
   */
  public void printlist(){
    System.out.print(this.first.toString());
    this.rest.printlist();
  }
  
  /**
   * Is this list sorted in the order given by the <code>Comparator</code>
   * @param comp the given <code>Comparator</code>
   * @return true if the list is sorted
   */
  public boolean isSorted(Comparator<T> comp){
    return this.rest.isSortedAfter(this.first, comp);
  }
  
  /**
   * Is this list sorted after the given item 
   * in the order given by the <code>Comparator</code>
   * @param item the given item that should precede the list
   * @param comp the given <code>Comparator</code>
   * @return true if the list is sorted
   */
  public boolean isSortedAfter(T item, Comparator<T> comp){
    return (comp.compare(item, this.first) <= 0) &&
           this.rest.isSortedAfter(this.first, comp);
  }
}