package sorting;
import java.util.*;

/**
 * One of the several sorting algorithms
 * to use for sorting detective work
 * @author Viera K. Proulx
 *
 * @param <T> the type of data to be sorted
 */
class SortingQuicksort<T> implements SortAlgorithm<T>{

  /** The <code>ArrayList</code> to sort */
  public ArrayList<T> alist;
  
  /**
   * Initialize the list to be sorted from the given data.
   * @param alist the <code>ArrayList</code> to sort
   */
  public void init(ArrayList<T> alistin){
    this.alist = new ArrayList<T>();
    for (T t: alistin){
      this.alist.add(t);
    }
  }
  
  /**
   * Sort the given <code>ArrayList</code> using the given 
   * <code>Comparator</code>. Return the sorted result.
   * For <em>in-place</em> algorithms <code>alist</code> should be
   * the result.
   * @param alist the <code>ArrayList</code> to sort
   * @param comp the <code>Comparator</code> that determines the ordering
   * @return the sorted <code>ArrayList</code>
   */
  public ArrayList<T> sort(Comparator<T> comp){
    quicksort(this.alist, comp);
    return this.alist;
  }
  

  // Return random Integer between: [low, high)
  public Integer randInt(int low, int high){ 
    return new Random().nextInt( high - low ) + low; 
  }
  
  // The swap we've seen many times in previous labs/hws
  void swap(ArrayList<T> list, int i, int j){
     list.set( i, list.set( j, list.get( i ) ) );
  }
     
  
  /**
   * Sort the given list using selection sort
   * @param <T> The type of data in the <code>ArrayList</code>
   * @param alistthe <code>ArrayList</code> to sort
   * @param comp the <code>Comparator</code> that determines the ordering
   * @return
   */
  public void quicksort(ArrayList<T> list, 
                            Comparator<T> comp){
    quicksortHelper( list, comp, 0, list.size()-1 );
  }
  
  // sort the list in the range from start to end (inclusive)
  void quicksortHelper(ArrayList<T> list, 
                           Comparator<T> comp, 
                           int start, int end){ 
    if ( end > start ){
      int pivot = partition( list, comp, start, end );

      quicksortHelper( list, comp, start, pivot - 1 );
      quicksortHelper( list, comp, pivot + 1, end );
    }  
  }
  
  //pick a random pivot and swap it into the first position
  T pickPivot(ArrayList<T> list, 
                  Comparator<T> comp, 
                  int start, int end){
    int pivot = randInt(start, end+1);
    T pivotVal = list.get(pivot);
    //System.out.println(pivotVal.toString());
    swap( list, start, pivot );
    return pivotVal;
  }

  // partition the list in the range from start to end (inclusive)
  int partition(ArrayList<T> list, 
                     Comparator<T> comp, 
                     int start, int end){
    
    // pick a random pivot and swap it into the start position
    T pivotVal = this.pickPivot(list, comp, start, end);
    
    // start on the left with the one position higher than the start position
    int left_swap_index = start+1;
    
    // start on the right with the given end
    int right_swap_index = end;
    
    // find the swap candidates and swap them
    // - stop when the middle is reached
    while ( left_swap_index < right_swap_index ) {
      
      // go up the list till the value is greater than the pivot
      // check not to go beyond the item already seen on the right
      while ( comp.compare( list.get( left_swap_index ),
                            pivotVal ) <= 0
                &&
              left_swap_index < right_swap_index )
        left_swap_index++;
      
      // go down the list till the value is less than or equal to the pivot
      // check not to go beyond the item already seen on the left
      while ( comp.compare( list.get( right_swap_index ), 
                            pivotVal ) > 0
                &&
              right_swap_index > left_swap_index )
        right_swap_index--;
      
      // swap the two values
      swap( list, left_swap_index, right_swap_index );
    }
    
    // left_swap_index and right_swap_index should be equal now
    // the mid point is either the lowest of the upper half
    // or the highest of the lower half
    int mid = left_swap_index;

    // make mid point to the highest of the lower half
    if ( comp.compare( list.get( mid ), pivotVal ) > 0 ) 
      mid = mid - 1;
    
    // now swap the pivot into its position
    swap( list, start, mid );
    
    return mid;
  }

}
    